﻿Imports System.Text

''' <summary>
''' This is the arbitrary precision Big Integer library.  It is somewhat analogous to the
''' System.Numerics.BigInteger library except here you can set the resultant precision 
''' (rather than have it automatically expand).  It also contains a few calculator-related
''' functions that do not appear in System.Numerics.BigInteger
''' </summary>
''' <remarks>Word sizes do not have to be divisible by 8</remarks>
Public Class BigInt
    Private pfvOverflow As Boolean
    Private pfvCarryBit As Boolean
    Private pfvLossOfPrecision As Boolean
    Private pfvArithMode As ArithMode

    Private BitSize As Integer
    Private LastByte As Integer
    Private n() As Byte

    ''' <summary>
    ''' The Arithmetic mode for signed decimal values
    ''' </summary>
    ''' <remarks>Warning: These values are synced with CalcState's CalcState.CalcArithMode</remarks>
    Public Enum ArithMode As Integer
        Unsigned
        OnesComp
        TwosComp
    End Enum

#Region "Constructors"

    ''' <summary>
    ''' The default constructor
    ''' </summary>
    ''' <remarks>Uses a default bit size of 32</remarks>
    Public Sub New()
        pfvOverflow = False
        pfvCarryBit = False
        pfvLossOfPrecision = False
        pfvArithMode = ArithMode.TwosComp
        BitSize = 32
        LastByte = 3
        ReDim n(LastByte)
    End Sub

    ''' <summary>
    ''' Constructor given a required size
    ''' </summary>
    ''' <param name="size">The bit size</param>
    ''' <param name="mode">The Arithmetic Mode</param>
    ''' <remarks></remarks>
    Public Sub New(size As Integer, mode As ArithMode)
        Me.New()
        pfvArithMode = mode

        SetSize(size)
    End Sub

    ''' <summary>
    ''' Constructor from a byte array
    ''' </summary>
    ''' <param name="byte_array">The byte array to convert</param>
    ''' <param name="size">The bit size</param>
    ''' <param name="mode">The Arithmetic Mode</param>
    ''' <remarks>Uses the bit size of the array</remarks>
    Public Sub New(byte_array As Byte(), size As Integer, mode As ArithMode)
        Me.New()
        pfvArithMode = mode

        Import(byte_array, size, False)
    End Sub

    ''' <summary>
    ''' Constructor from another BigInt
    ''' </summary>
    ''' <param name="bi">The BigInt to convert</param>
    ''' <param name="size">The bit size</param>
    ''' <param name="mode">The Arithmetic Mode</param>
    ''' <remarks></remarks>
    Public Sub New(bi As BigInt, size As Integer, mode As ArithMode)
        Me.New()
        pfvArithMode = mode

        Import(bi.n, size, False)
    End Sub

    ''' <summary>
    ''' Constructor from a double
    ''' </summary>
    ''' <param name="dbl">The double to convert</param>
    ''' <param name="size">The bit size</param>
    ''' <param name="mode">The Arithmetic Mode</param>
    ''' <remarks>Resulting bit size is 16</remarks>
    Public Sub New(dbl As Double, size As Integer, mode As ArithMode)
        Me.New()
        pfvArithMode = mode

        Dim bytes() As Byte
        Dim temp As Long

        ' We actually just convert the double to a long, and then
        ' perform the normal import for a long.

        ' Convert only if there is some useful value
        If dbl >= Long.MaxValue Or dbl <= Long.MinValue Then
            ' Humm... should this generate an exception?
            pfvLossOfPrecision = True
            Exit Sub
        End If

        ' v7.0.0 - 6 Jun 21: Fix 1's complement error
        If dbl < 0 And mode = ArithMode.OnesComp Then
            dbl -= 1
        End If

        ' Use default rounding rules!
        temp = CLng(dbl)
        bytes = BitConverter.GetBytes(temp)
        If Not BitConverter.IsLittleEndian Then
            Array.Reverse(bytes)
        End If

        ' only 15 digit precision for a double
        Import(bytes, size, (dbl < 0))
    End Sub

    ''' <summary>
    ''' Constructor from an integer
    ''' </summary>
    ''' <param name="i">The integer to convert</param>
    ''' <param name="size">The bit size</param>
    ''' <param name="mode">The Arithmetic Mode</param>
    ''' <remarks>Negative numbers imported as 2's complement, then the mode is applied</remarks>
    Public Sub New(i As Int32, size As Integer, mode As ArithMode)
        Me.New()
        pfvArithMode = mode

        Dim bytes() As Byte

        ' v7.0.0 - 6 Jun 21: Fix 1's complement error
        If i < 0 And mode = ArithMode.OnesComp Then
            i -= 1
        End If

        bytes = BitConverter.GetBytes(i)
        If Not BitConverter.IsLittleEndian Then
            Array.Reverse(bytes)
        End If

        Import(bytes, size, (i < 0))
    End Sub

    ''' <summary>
    ''' Constructor from a long
    ''' </summary>
    ''' <param name="lng">The long to convert</param>
    ''' <param name="size">The bit size</param>
    ''' <param name="mode">The Arithmetic Mode</param>
    ''' <remarks>Negative numbers imported as 2's complement, then the mode is applied</remarks>
    Public Sub New(lng As Int64, size As Integer, mode As ArithMode)
        Me.New()
        pfvArithMode = mode

        Dim bytes() As Byte

        ' v7.0.0 - 6 Jun 21: Fix 1's complement error
        If lng < 0 And mode = ArithMode.OnesComp Then
            lng -= 1
        End If

        bytes = BitConverter.GetBytes(lng)
        If Not BitConverter.IsLittleEndian Then
            Array.Reverse(bytes)
        End If

        Import(bytes, size, (lng < 0))
    End Sub

    ''' <summary>
    ''' Constructor from an unsigned integer
    ''' </summary>
    ''' <param name="ui">The unsigned integer to convert</param>
    ''' <param name="size">The bit size</param>
    ''' <param name="mode">The Arithmetic Mode</param>
    ''' <remarks></remarks>
    Public Sub New(ui As UInt32, size As Integer, mode As ArithMode)
        Me.New()
        pfvArithMode = mode

        Dim bytes() As Byte

        bytes = BitConverter.GetBytes(ui)
        If Not BitConverter.IsLittleEndian Then
            Array.Reverse(bytes)
        End If

        Import(bytes, size, False)
    End Sub

    ''' <summary>
    ''' Constructor from an unsigned long
    ''' </summary>
    ''' <param name="ul">The unsigned long to convert</param>
    ''' <param name="size">The bit size</param>
    ''' <param name="mode">The Arithmetic Mode</param>
    ''' <remarks></remarks>
    Public Sub New(ul As UInt64, size As Integer, mode As ArithMode)
        Me.New()
        pfvArithMode = mode

        Dim bytes() As Byte

        bytes = BitConverter.GetBytes(ul)
        If Not BitConverter.IsLittleEndian Then
            Array.Reverse(bytes)
        End If

        Import(bytes, size, False)
    End Sub

    ''' <summary>
    ''' Constructor from a string
    ''' </summary>
    ''' <param name="s">The string to convert</param>
    ''' <param name="size">The bit size</param>
    ''' <param name="mode">The Arithmetic Mode</param>
    ''' <remarks>
    ''' Uses a "&amp;b" notation for binary. Formatting errors are not
    ''' processed here
    ''' </remarks>
    Public Sub New(s As String, size As Integer, mode As ArithMode)
        Me.New()

        Dim i, bits, temp As Integer
        Dim n_byte, n_bit, finish As Integer
        Dim c, number As String
        Dim z As BigInt

        If String.IsNullOrEmpty(s) Then
            Exit Sub
        End If

        ' v7.0.1 - 3 Nov 21: Let's simplify calculation of the size
        z = New BigInt(size + 8, mode)
        Me.SetSize(size + 8)
        pfvArithMode = mode
        pfvLossOfPrecision = False

        If s.Length > 2 AndAlso s.StartsWith("&") Then
            number = Trim(s.Substring(2).ToLower)
            Select Case s.Substring(1, 1).ToLower
                Case "o"    ' Octal notation
                    finish = 0
                    ' input validation from right to left...
                    For i = number.Length - 1 To 0 Step -1
                        ' Stop at first unknown character
                        c = number.Substring(i, 1)
                        If c.IndexOfAny("01234567".ToCharArray) = -1 Then
                            finish = i + 1
                            Exit For
                        End If
                    Next
                    ' Process the string
                    For i = number.Length - 1 To finish Step -1
                        temp = CInt(number.Substring(i, 1))
                        If temp <> 0 Then
                            ' Sneaky trick to prevent multiplication/shifts
                            z.Clear()
                            z.PutAt(temp, bits)
                            _Add(z)
                        End If
                        ' Number of bits to shift
                        bits = bits + 3
                        ' v7.0.2 - 28 Jun 23: Trimming input during shortening bit size
                        If bits > size Then
                            Exit For
                        End If
                    Next
                Case "b"    ' Binary (not a really a standard notation!)
                    finish = 0
                    ' input validation from right to left...
                    For i = number.Length - 1 To 0 Step -1
                        ' Stop at first unknown character
                        c = number.Substring(i, 1)
                        If c.IndexOfAny("01".ToCharArray) = -1 Then
                            finish = i + 1
                            Exit For
                        End If
                    Next
                    ' v7.0.2 - 28 Jun 23: Trimming input during shortening bit size
                    If finish > size Then
                        finish = size
                    End If
                    ' Process the string
                    For i = number.Length - 1 To finish Step -1
                        ' Which byte and which bit?
                        n_byte = (number.Length - 1 - i) \ 8
                        n_bit = (number.Length - 1 - i) Mod 8
                        If number.Substring(i, 1) = "1" Then
                            Me.n(n_byte) = Me.n(n_byte) + CByte(2 ^ n_bit)
                        End If
                    Next
                Case "h"    ' Hexadecimal notation
                    finish = 0
                    ' input validation from right to left...
                    For i = number.Length - 1 To 0 Step -1
                        ' Stop at first unknown character
                        c = number.Substring(i, 1)
                        If c.IndexOfAny("0123456789abcdef".ToCharArray) = -1 Then
                            finish = i + 1
                            Exit For
                        End If
                    Next

                    ' Process the string
                    For i = number.Length - 1 To finish Step -1
                        temp = CInt("&H" & number.Substring(i, 1))
                        If temp <> 0 Then
                            ' Sneaky trick to prevent multiplication/shifts
                            z.Clear()
                            z.PutAt(temp, bits)
                            _Add(z)
                        End If
                        ' Number of bits to shift
                        bits = bits + 4
                        ' v7.0.2 - 28 Jun 23: Trimming input during shortening bit size
                        If bits > size Then
                            Exit For
                        End If
                    Next
            End Select
        Else
            ' If no leading "&", it must be a decimal notation
            finish = 0
            ' input validation from right to left.  This might seem a bit counterintuitive
            ' since "123.45" would yield "45" and not "123"
            For i = s.Length - 1 To 0 Step -1
                ' Stop at first unknown character (including a decimal or negative sign)
                c = s.Substring(i, 1)
                If c.IndexOfAny("0123456789".ToCharArray) = -1 Then
                    finish = i + 1
                    Exit For
                End If
            Next

            Dim mult As New BigInt(size + 8, mode)
            Dim ten As New BigInt(size + 8, mode)
            Dim one As BigInt = BigInt.One(size + 8)

            ' This is normal stuff.  You multiply each digit times the value
            ' associated with that decimal place, then add 'em all up.
            mult.n(0) = 1
            ten.n(0) = 10
            For i = s.Length - 1 To finish Step -1
                temp = CInt(s.Substring(i, 1))
                If temp <> 0 Then
                    z.Clear()
                    z.n(0) = CByte(temp)
                    ' Can't use my sneaky PutAt function...
                    z.Multiply(mult, BitSize)
                    _Add(z)
                End If
                ' Update the Decimal place
                mult.Multiply(ten, BitSize)
                ' Did you overflow the mult?
                If mult.Overflow Then
                    pfvLossOfPrecision = True
                    Exit For
                End If
            Next

            ' If it's negative assume 2's complement
            If s.StartsWith("-") Then
                OnesComplement(BitSize)
                ' v7.0.0 - 6 Jun 21: Fix 1's complement error
                If mode = ArithMode.TwosComp Then
                    _Add(one)
                End If

            End If
        End If

        ' We need to know if the input has exceeded the capacity of 
        ' the word size.  If so, then use the LossOfPrecision flag
        ' so that ProcessKey can remove the last digit entered.

        z = Me.Copy
        SetSize(size)
        ' v7.0.1 - 3 Nov 21: Fixed a numeric entry "sign extension" bug
        If z.MaxBit <> Me.MaxBit Then
            pfvLossOfPrecision = True
        End If

        ' v7.0.0 - 6 Jun 21: Fix 1's complement error
        ' manually extend the sign
        If s.StartsWith("-") AndAlso mode = ArithMode.Unsigned Then
            For ext As Integer = LastByte To 0 + 1
                If n(ext) = 0 Then
                    n(ext) = CByte(255)
                Else
                    Exit For
                End If
            Next
        End If
    End Sub
#End Region

#Region "Properties"

    ''' <summary>
    ''' Math overflow condition
    ''' </summary>
    ''' <value></value>
    ''' <returns>True if an overflow has occurred</returns>
    ''' <remarks></remarks>
    Public ReadOnly Property Overflow As Boolean
        Get
            Return pfvOverflow
        End Get
    End Property

    ''' <summary>
    ''' Carry Bit
    ''' </summary>
    ''' <value></value>
    ''' <returns>True if the carry bit is set</returns>
    ''' <remarks></remarks>
    Public ReadOnly Property CarryBit As Boolean
        Get
            Return pfvCarryBit
        End Get
    End Property

    ''' <summary>
    ''' Loss of Precision
    ''' </summary>
    ''' <value></value>
    ''' <returns>Returns true if an assignment to a BigInt has loss precision</returns>
    ''' <remarks></remarks>
    Public ReadOnly Property LossOfPrecision As Boolean
        Get
            Return pfvLossOfPrecision
        End Get
    End Property

    ''' <summary>
    ''' Get or Set the Arithmetic (complement) mode
    ''' </summary>
    ''' <value>The new Arithmetic mode</value>
    ''' <returns>The current Arithmetic mode</returns>
    ''' <remarks></remarks>
    Public Property BIArithMode As ArithMode
        Get
            Return pfvArithMode
        End Get
        Set(value As ArithMode)
            pfvArithMode = value
        End Set
    End Property

    ''' <summary>
    ''' Get or Set the word size in bits
    ''' </summary>
    ''' <value>The new word size</value>
    ''' <returns>The current word size</returns>
    ''' <remarks>A reduction is size does not trigger an Loss Of Precision</remarks>
    Public Property WordSize As Integer
        Get
            Return BitSize
        End Get
        Set(value As Integer)
            SetSize(value)
            BitSize = value
        End Set
    End Property

    ''' <summary>
    ''' Direct access to the underlying data
    ''' </summary>
    ''' <value>The new byte array</value>
    ''' <returns>The existing byte array</returns>
    ''' <remarks>No error checking</remarks>
    Public Property ByteArray As Byte()
        Get
            Return n
        End Get
        Set(value As Byte())
            n = value
        End Set
    End Property

#End Region

#Region "Math Operators"

    ''' <summary>
    ''' Add operator
    ''' </summary>
    ''' <param name="x">Left value</param>
    ''' <param name="y">Right value</param>
    ''' <returns>A new BigInt with the results of the math operation</returns>
    ''' <remarks></remarks>
    Public Shared Operator +(x As BigInt, y As BigInt) As BigInt
        Dim ans As BigInt = x.Copy
        ans.Add(y, 0)
        Return ans
    End Operator

    ''' <summary>
    ''' Add two BigInt values
    ''' </summary>
    ''' <param name="x">Left value</param>
    ''' <param name="y">Right value</param>
    ''' <returns>A new BigInt with the results of the math operation</returns>
    ''' <remarks>Shared function</remarks>
    Public Shared Function Add(x As BigInt, y As BigInt) As BigInt
        Dim ans As BigInt = x.Copy
        ans.Add(y, 0)
        Return ans
    End Function

    ''' <summary>
    ''' Add two BigInt values with required size
    ''' </summary>
    ''' <param name="x">Left value</param>
    ''' <param name="y">Right value</param>
    ''' <param name="size">The required size of the resultant BigInt</param>
    ''' <returns>A new BigInt with the results of the math operation</returns>
    ''' <remarks>Shared function</remarks>
    Public Shared Function Add(x As BigInt, y As BigInt, size As Integer) As BigInt
        Dim ans As BigInt = x.Copy
        ans.Add(y, size)
        Return ans
    End Function

    ''' <summary>
    ''' Add a BigInt value to the current value
    ''' </summary>
    ''' <param name="y">The value to add to the current instance</param>
    ''' <param name="size">The required bit size of the resultant</param>
    ''' <remarks></remarks>
    Public Sub Add(ByVal y As BigInt, ByVal size As Integer)
        Dim real_bitsize As Integer
        Dim temp_y As BigInt
        Dim sign_x, sign_y As Boolean

        temp_y = y.Copy

        ' determine the bitsize, and make 'em all the same
        real_bitsize = If(size > 0, size, If(BitSize >= temp_y.BitSize, BitSize, temp_y.BitSize))
        SetSize(real_bitsize)
        temp_y.SetSize(real_bitsize)

        sign_x = Me.SignBit
        sign_y = temp_y.SignBit

        ' We temporarily need a bit more "wiggle room" than just the current bitsize,
        ' so we add another byte.  This is similar to SetSize, but never does a
        ' sign extension
        AddPadding(RoundUp(real_bitsize))
        temp_y.AddPadding(BitSize)

        _Add(temp_y)

        ' We needed the wiggle room to detect carry and overflow
        pfvCarryBit = MaxBit() >= real_bitsize
        Select Case BIArithMode
            Case ArithMode.Unsigned
                ' that was easy...
                pfvOverflow = pfvCarryBit
            Case ArithMode.OnesComp
                ' the "wrap around" rule for adding 1's complement numbers
                If pfvCarryBit Then
                    _Add(BigInt.One(BitSize))
                End If

                ' determining overflow is a bit bizarre
                pfvOverflow = False
                If sign_x Then
                    If sign_y And Not Me.BitTest(real_bitsize - 1) And pfvCarryBit Then
                        pfvOverflow = True
                    End If
                Else
                    If Not sign_y And Me.BitTest(real_bitsize - 1) And Not pfvCarryBit Then
                        pfvOverflow = True
                    End If
                End If
            Case ArithMode.TwosComp
                ' determining overflow is a bit bizarre
                pfvOverflow = False
                If sign_x Then
                    If sign_y And Not Me.BitTest(real_bitsize - 1) And pfvCarryBit Then
                        pfvOverflow = True
                    End If
                Else
                    If Not sign_y And Me.BitTest(real_bitsize - 1) And Not pfvCarryBit Then
                        pfvOverflow = True
                    End If
                End If
        End Select

        ' OK, but it back to the real bitsize
        SetSize(real_bitsize)
    End Sub

    ''' <summary>
    ''' Internal version of Add (does not set Carry, Overflow, LOP)
    ''' </summary>
    ''' <param name="y">The value to add to the current instance</param>
    ''' <remarks></remarks>
    Private Sub _Add(y As BigInt)
        Dim carry, temp As Integer

        ' Normal stuff... add two bytes, keep the "carry" for the next byte
        carry = 0
        For i As Integer = 0 To Me.LastByte
            temp = CInt(Me.n(i)) + CInt(y.n(i)) + carry
            Me.n(i) = CByte(temp Mod 256)
            carry = temp \ 256
        Next

        Mask()
    End Sub

    ''' <summary>
    ''' Substract operator
    ''' </summary>
    ''' <param name="x">Left value</param>
    ''' <param name="y">Right value</param>
    ''' <returns>A new BigInt with the results of the math operation</returns>
    ''' <remarks></remarks>
    Public Shared Operator -(x As BigInt, y As BigInt) As BigInt
        Dim ans As BigInt = x.Copy
        ans.Subtract(y, 0)
        Return ans
    End Operator

    ''' <summary>
    ''' Subtract two BigInt values
    ''' </summary>
    ''' <param name="x">Left value</param>
    ''' <param name="y">Right value</param>
    ''' <returns>A new BigInt with the results of the math operation</returns>
    ''' <remarks>Shared function</remarks>
    Public Shared Function Subtract(x As BigInt, y As BigInt) As BigInt
        Dim ans As BigInt = x.Copy
        ans.Subtract(y, 0)
        Return ans
    End Function

    ''' <summary>
    ''' Subtract two BigInt values with required size
    ''' </summary>
    ''' <param name="x">Left value</param>
    ''' <param name="y">Right value</param>
    ''' <param name="size">The required size of the resultant BigInt</param>
    ''' <returns>A new BigInt with the results of the math operation</returns>
    ''' <remarks>Shared function</remarks>
    Public Shared Function Subtract(x As BigInt, y As BigInt, size As Integer) As BigInt
        Dim ans As BigInt = x.Copy
        ans.Subtract(y, size)
        Return ans
    End Function

    ''' <summary>
    ''' Subtract a BigInt value from the current value
    ''' </summary>
    ''' <param name="y">The value to subtract</param>
    ''' <param name="size">The required bit size of the resultant</param>
    ''' <remarks></remarks>
    Public Sub Subtract(y As BigInt, size As Integer)
        Dim real_bitsize As Integer
        Dim temp_y As BigInt
        Dim sign_x, sign_y As Boolean

        temp_y = y.Copy

        ' determine the bitsize, and make 'em all the same
        real_bitsize = If(size > 0, size, If(BitSize >= temp_y.BitSize, BitSize, temp_y.BitSize))
        SetSize(real_bitsize)
        temp_y.SetSize(real_bitsize)

        sign_x = Me.SignBit
        sign_y = temp_y.SignBit

        ' We temporarily need a bit more "wiggle room" than just the current bitsize,
        ' so we add another byte.  This is similar to SetSize, but never does a
        ' sign extension
        AddPadding(RoundUp(real_bitsize))
        temp_y.AddPadding(BitSize)

        _Subtract(temp_y)

        ' We needed the wiggle room to detect carry and overflow
        pfvCarryBit = MaxBit() >= real_bitsize
        Select Case BIArithMode
            Case ArithMode.Unsigned
                ' that was easy...
                pfvOverflow = pfvCarryBit
            Case ArithMode.OnesComp
                ' the "wrap around" rule for subtracting 1's complement numbers
                If pfvCarryBit Then
                    _Subtract(BigInt.One(BitSize))
                End If

                ' determining overflow is a bit bizarre
                pfvOverflow = False
                If sign_x Then
                    If Not sign_y And Not Me.BitTest(real_bitsize - 1) And Not pfvCarryBit Then
                        pfvOverflow = True
                    End If
                Else
                    If sign_y And Me.BitTest(real_bitsize - 1) And pfvCarryBit Then
                        pfvOverflow = True
                    End If
                End If
            Case ArithMode.TwosComp
                ' determining overflow is a bit bizarre
                pfvOverflow = False
                If sign_x Then
                    If Not sign_y And Not Me.BitTest(real_bitsize - 1) And Not pfvCarryBit Then
                        pfvOverflow = True
                    End If
                Else
                    If sign_y And Me.BitTest(real_bitsize - 1) And pfvCarryBit Then
                        pfvOverflow = True
                    End If
                End If
        End Select

        ' OK, but it back to the real bitsize
        SetSize(real_bitsize)
    End Sub

    ''' <summary>
    ''' Internal version of Subtract (does not set Carry, Overflow)
    ''' </summary>
    ''' <param name="y">The value to subtract</param>
    ''' <remarks></remarks>
    Private Sub _Subtract(y As BigInt)
        Dim carry, temp As Integer

        ' Normal stuff... subtract two bytes, generate a "borrow" if needed
        carry = 0
        For i As Integer = 0 To Me.LastByte
            temp = CInt(Me.n(i)) - CInt(y.n(i)) - carry
            ' Do we need a "borrow"?
            If temp < 0 Then
                temp = temp + 256
                carry = 1
            Else
                carry = 0
            End If
            Me.n(i) = CByte(temp)
        Next

        Mask()
    End Sub

    ''' <summary>
    ''' Multiply operator
    ''' </summary>
    ''' <param name="x">Left value</param>
    ''' <param name="y">Right value</param>
    ''' <returns>A new BigInt with the results of the math operation</returns>
    ''' <remarks></remarks>
    Public Shared Operator *(x As BigInt, y As BigInt) As BigInt
        Dim ans As BigInt = x.Copy
        ans.Multiply(y, 0)
        Return ans
    End Operator

    ''' <summary>
    ''' Multiply two BigInt values
    ''' </summary>
    ''' <param name="x">Left value</param>
    ''' <param name="y">Right value</param>
    ''' <returns>A new BigInt with the results of the math operation</returns>
    ''' <remarks>Shared function</remarks>
    Public Shared Function Multiply(x As BigInt, y As BigInt) As BigInt
        Dim ans As BigInt = x.Copy
        ans.Multiply(y, 0)
        Return ans
    End Function

    ''' <summary>
    ''' Multiply two BigInt values with required size
    ''' </summary>
    ''' <param name="x">Left value</param>
    ''' <param name="y">Right value</param>
    ''' <param name="size">The size of the resultant BigInt</param>
    ''' <returns>A new BigInt with the results of the math operation</returns>
    ''' <remarks>Shared function</remarks>
    Public Shared Function Multiply(x As BigInt, y As BigInt, size As Integer) As BigInt
        Dim ans As BigInt = x.Copy
        ans.Multiply(y, size)
        Return ans
    End Function

    ''' <summary>
    ''' Multiply a BigInt value with the current value
    ''' </summary>
    ''' <param name="y">The value to multiply</param>
    ''' <param name="size">The required bit size of the resultant</param>
    ''' <remarks></remarks>
    Public Sub Multiply(ByVal y As BigInt, ByVal size As Integer)
        Dim temp, real_bitsize As Integer
        Dim temp_x, temp_y, z As BigInt
        Dim neg As Boolean = False

        temp_x = Me.Copy
        temp_y = y.Copy

        ' We need a bit more "wiggle room" than just the current bitsize,
        ' so round off to the nearest byte and add 1 byte
        real_bitsize = If(size > 0, size, If(BitSize >= temp_y.BitSize, BitSize, temp_y.BitSize))
        SetSize(RoundUp(real_bitsize))
        temp_x.SetSize(BitSize)
        temp_y.SetSize(BitSize)
        z = New BigInt(BitSize, pfvArithMode)
        Clear()

        ' Let's handle some special cases
        If temp_x.IsZero() Or temp_y.IsZero Then
            SetSize(real_bitsize)
            Exit Sub
        End If

        ' v7.0.0 - 6 Jun 21: Fix 1's complement error
        If pfvArithMode <> ArithMode.Unsigned Then
            If temp_x.SignBit() Xor temp_y.SignBit() Then
                neg = True
            End If
            temp_x.AbsoluteValue()
            temp_y.AbsoluteValue()
        End If

        ' This is just like you were taught is school...  multiply each digit
        ' in the first number with every digit in the second.  Keep track of
        ' what "column" to write down the intermediate values, then add the
        ' intermediate values together.

        For i As Integer = 0 To temp_x.MaxByte()
            For j As Integer = 0 To temp_y.MaxByte()
                temp = CInt(temp_x.n(i)) * CInt(temp_y.n(j))
                ' anything to do?
                If temp <> 0 Then
                    If (i + j) <= LastByte Then
                        ' This is a sneaky trick to create intermediate values
                        ' without using shifts.  Normally you'd multiply the
                        ' temp value times the value for that "decimal place".
                        z.Clear()
                        z.PutAt(temp, (i + j) * 8)
                        _Add(z)
                    Else
                        pfvOverflow = True
                    End If
                End If
            Next
        Next
        ' We needed the wiggle room to detect over_flow
        If Not pfvOverflow Then
            pfvOverflow = MaxBit() >= real_bitsize
        End If

        SetSize(real_bitsize)

        If neg Then
            ChangeSign()
        End If
    End Sub

    ''' <summary>
    ''' Divide operator
    ''' </summary>
    ''' <param name="x">Dividend (Left value)</param>
    ''' <param name="y">Divisor (Right value)</param>
    ''' <returns>A new BigInt with the results of the math operation</returns>
    ''' <remarks></remarks>
    Public Shared Operator /(x As BigInt, y As BigInt) As BigInt
        Dim ans As BigInt = x.Copy
        ans.Divide(y, 0)
        Return ans
    End Operator

    ''' <summary>
    ''' Divide two BigInt values
    ''' </summary>
    ''' <param name="x">Dividend (Left value)</param>
    ''' <param name="y">Divisor (Right value)</param>
    ''' <returns>A new BigInt with the results of the math operation</returns>
    ''' <remarks>Shared function</remarks>
    Public Shared Function Divide(x As BigInt, y As BigInt) As BigInt
        Dim ans As BigInt = x.Copy
        ans.Divide(y, 0)
        Return ans
    End Function

    ''' <summary>
    ''' Divide two BigInt values with required size
    ''' </summary>
    ''' <param name="x">Dividend (Left value)</param>
    ''' <param name="y">Divisor (Right value)</param>
    ''' <param name="size">The required bit size of the resultant</param>
    ''' <returns>A new BigInt with the results of the math operation</returns>
    ''' <remarks>Shared function</remarks>
    Public Shared Function Divide(x As BigInt, y As BigInt, size As Integer) As BigInt
        Dim ans As BigInt = x.Copy
        ans.Divide(y, size)
        Return ans
    End Function


    ''' <summary>
    ''' Divide the current value with a BigInt value
    ''' </summary>
    ''' <param name="y">The value to Divide by</param>
    ''' <param name="size">The required bit size of the resultant</param>
    ''' <remarks></remarks>
    Public Sub Divide(ByVal y As BigInt, ByVal size As Integer)
        Dim nx, ny, nc, ans_digit, real_bitsize As Integer
        Dim chunk, guess, temp_x, temp_y, z As BigInt
        Dim neg As Boolean = False

        temp_x = Me.Copy
        temp_y = y.Copy

        ' We need a bit more "wiggle room" than just the current bitsize,
        ' so round off to the nearest byte and add 1 byte
        real_bitsize = If(size > 0, size, If(BitSize >= temp_y.BitSize, BitSize, temp_y.BitSize))
        SetSize(RoundUp(real_bitsize))
        temp_x.SetSize(BitSize)
        temp_y.SetSize(BitSize)
        guess = New BigInt(BitSize, pfvArithMode)
        chunk = New BigInt(BitSize, pfvArithMode)

        ' Sanity check
        If temp_y.IsZero() Then
            pfvOverflow = True
            SetSize(real_bitsize)
            Clear()
            Exit Sub
        End If

        ' When using signed numbers, we first figure out the what the sign of 
        ' the answer will be, then convert everything to absolute value
        If pfvArithMode <> ArithMode.Unsigned Then
            ' if either is negative (but not both)
            If temp_x.SignBit Xor temp_y.SignBit Then
                neg = True
            End If

            temp_x.AbsoluteValue()
            temp_y.AbsoluteValue()
        End If

        ' Special case
        If temp_x.IsLess(temp_y) Then
            If Not temp_x.IsZero() Then
                pfvCarryBit = True
            End If
            SetSize(real_bitsize)
            Clear()
            Exit Sub
        End If

        ' This is almost like you were taught in school... For each digit in
        ' the answer, guess the number of times the divisor will go into a
        ' "chunk" of the dividend.  We know the size of the chunk will have
        ' the same number of digits as x or be one digit larger.  Subtract
        ' the intermediate value and repeat.

        Clear()
        ny = temp_y.MaxByte()
        Do
            ' If no more divisors... we're done
            If temp_x.IsLess(temp_y) Then
                Exit Do
            End If

            nx = temp_x.MaxByte()
            ' Build a "chunk" that has the same number of digits as y
            chunk.Clear()
            For i As Integer = 0 To ny
                chunk.n(i) = temp_x.n(i + nx - ny)
            Next
            nc = ny

            ' If it is too small, build another chunk with one more digit
            If chunk.IsLess(temp_y) Then
                For i As Integer = 0 To ny + 1
                    chunk.n(i) = temp_x.n(i + nx - ny - 1)
                Next
                nc = ny + 1
            End If

            ' To keep from testing all 255 possibilities for each "answer
            ' digit", we use a simplified binary tree search algorithm.
            ans_digit = 128
            guess.Clear()
            For i As Integer = 6 To 0 Step -1
                guess.n(0) = CByte(ans_digit)
                ' This is why we need more wiggle room... so we don't have 
                ' to deal with overflows during the guessing of the "answer
                ' digit"
                z = Multiply(temp_y, guess)
                If z.IsGreater(chunk) Then
                    ans_digit = CInt(ans_digit - (2 ^ i))
                Else
                    ans_digit = CInt(ans_digit + (2 ^ i))
                End If
            Next
            guess.n(0) = CByte(ans_digit)
            z = Multiply(temp_y, guess)
            If z.IsGreater(chunk) Then
                ans_digit = ans_digit - 1
            End If

            ' Place answer digit in the proper "column"
            z.Clear()
            z.PutAt(ans_digit, (nx - nc) * 8)
            ' Add it to the answer.
            _Add(z)

            ' Generate the intermediate value by shifting the "last guess"
            ' multiplier to its proper place (which undoes the effect of the
            ' shift that occurred when we developed the chunk).
            guess.Clear()
            guess.n(nx - nc) = CByte(ans_digit)
            z = Multiply(temp_y, guess)

            ' Now subtract the intermediate value and continue...
            temp_x._Subtract(z)
        Loop While nx - nc > 0

        pfvCarryBit = Not temp_x.IsZero()
        pfvOverflow = MaxBit() >= real_bitsize

        SetSize(real_bitsize)

        If neg Then
            ChangeSign()
        End If
    End Sub

    ''' <summary>
    ''' Modulus operator
    ''' </summary>
    ''' <param name="x">Dividend (Left value)</param>
    ''' <param name="y">Divisor (Right value)</param>
    ''' <returns>A new BigInt with the results of the math operation</returns>
    ''' <remarks></remarks>
    Public Shared Operator Mod(x As BigInt, y As BigInt) As BigInt
        Return Remainder(x, y, 0)
    End Operator

    ''' <summary>
    ''' Modulus (Division Remainder)
    ''' </summary>
    ''' <param name="x">Dividend (Left value)</param>
    ''' <param name="y">Divisor (Right value)</param>
    ''' <returns>A new BigInt with the results of the math operation</returns>
    ''' <remarks>There is no instance version</remarks>
    Public Shared Function [Mod](ByVal x As BigInt, ByVal y As BigInt) As BigInt
        Return Remainder(x, y, 0)
    End Function

    ''' <summary>
    ''' Modulus (Division Remainder) with required size
    ''' </summary>
    ''' <param name="x">Dividend (Left value)</param>
    ''' <param name="y">Divisor (Right value)</param>
    ''' <returns>A new BigInt with the results of the math operation</returns>
    ''' <remarks>There is no instance version</remarks>
    Public Shared Function [Mod](ByVal x As BigInt, ByVal y As BigInt, size As Integer) As BigInt
        Return Remainder(x, y, size)
    End Function

    ''' <summary>
    ''' Modulus (Division Remainder)
    ''' </summary>
    ''' <param name="x">Dividend (Left value)</param>
    ''' <param name="y">Divisor (Right value)</param>
    ''' <returns>A new BigInt with the results of the math operation</returns>
    ''' <remarks>There is no instance version</remarks>
    Public Shared Function Remainder(ByVal x As BigInt, ByVal y As BigInt) As BigInt
        Return Remainder(x, y, 0)
    End Function

    ''' <summary>
    ''' Modulus (Division Remainder) with required size
    ''' </summary>
    ''' <param name="x">Dividend (Left value)</param>
    ''' <param name="y">Divisor (Right value)</param>
    ''' <param name="size">The required bit size of the resultant value</param>
    ''' <returns>A new BigInt with the results of the math operation</returns>
    ''' <remarks>There is no instance version</remarks>
    Shared Function Remainder(ByVal x As BigInt, ByVal y As BigInt, ByVal size As Integer) As BigInt
        Dim nx, ny, nc, ans_digit, real_bitsize As Integer
        Dim chunk, guess As BigInt
        Dim temp_x, temp_y, z, ans As BigInt
        Dim neg As Boolean = False

        temp_x = x.Copy
        temp_y = y.Copy

        ' We need a bit more wiggle room than the just current bitsize.
        real_bitsize = If(size > 0, size, If(temp_x.BitSize >= temp_y.BitSize, temp_x.BitSize, temp_y.BitSize))
        ans = New BigInt(RoundUp(real_bitsize), x.pfvArithMode)
        temp_x.SetSize(ans.BitSize)
        temp_y.SetSize(ans.BitSize)
        guess = New BigInt(ans.BitSize, x.pfvArithMode)
        chunk = New BigInt(ans.BitSize, x.pfvArithMode)

        ' Sanity check
        If temp_y.IsZero Then
            ans.SetSize(real_bitsize)
            ans.Clear()
            ans.pfvOverflow = True
            Return ans
        End If

        ' When using signed numbers, we first figure out the what the sign of 
        ' the answer will be, then convert everything to absolute value
        If x.pfvArithMode <> ArithMode.Unsigned Then
            If x.SignBit Then
                neg = True
            End If

            temp_x.AbsoluteValue()
            temp_y.AbsoluteValue()
        End If

        ' Special case (if x < y, then return x)
        If temp_x.IsLess(temp_y) Then
            temp_x.SetSize(real_bitsize)
            If neg Then
                temp_x.pfvArithMode = ArithMode.TwosComp
                temp_x.ChangeSign()
                temp_x.pfvArithMode = x.pfvArithMode
            End If
            Return temp_x
        End If

        ' This is almost like you were taught in school... For each digit in
        ' the answer, guess the number of times the divisor will go into a
        ' "chunk" of the dividend.  We know the size of the chunk will have
        ' the same number of digits as x or be one digit larger.  Subtract
        ' the intermediate value and repeat.

        ans = temp_x.Copy
        ny = temp_y.MaxByte()
        Do
            ' If no more divisors... we're done
            If ans.IsLess(temp_y) Then
                Exit Do
            End If

            nx = ans.MaxByte()
            ' Build a "chunk" that has the same number of digits as y
            chunk.Clear()
            For i As Integer = 0 To ny
                chunk.n(i) = ans.n(i + nx - ny)
            Next
            nc = ny

            ' If it is too small, build another chunk with one more digit
            If chunk.IsLess(temp_y) Then
                For i As Integer = 0 To ny + 1
                    chunk.n(i) = ans.n(i + nx - ny - 1)
                Next
                nc = ny + 1
            End If

            ' To keep from testing all 255 possibilities for each "answer
            ' digit", we use a simplified binary tree search algorithm.
            ans_digit = 128
            guess.Clear()
            For i As Integer = 6 To 0 Step -1
                guess.n(0) = CByte(ans_digit)
                ' This is why we need more wiggle room... so we don't have 
                ' to deal with overflows during the guessing of the "answer
                ' digit"
                z = Multiply(temp_y, guess)
                If z.IsGreater(chunk) Then
                    ans_digit = CInt(ans_digit - (2 ^ i))
                Else
                    ans_digit = CInt(ans_digit + (2 ^ i))
                End If
            Next
            guess.n(0) = CByte(ans_digit)
            z = Multiply(temp_y, guess)
            If z.IsGreater(chunk) Then
                ans_digit = ans_digit - 1
            End If

            ' Generate the intermediate value by shifting the "last guess"
            ' multiplier to its proper place (which undoes the effect of the
            ' shift that occurred when we developed the chunk).
            guess.Clear()
            guess.n(nx - nc) = CByte(ans_digit)
            z = Multiply(temp_y, guess)

            ' Now subtract the intermediate value and continue...
            ans._Subtract(z)
        Loop While nx - nc > 0

        ans.SetSize(real_bitsize)
        If neg Then
            ans.ChangeSign()
        End If

        Return ans
    End Function

    ''' <summary>
    ''' Power Operator
    ''' </summary>
    ''' <param name="x">The base value</param>
    ''' <param name="y">The Exponent</param>
    ''' <returns>A new BigInt with the results of the math operation</returns>
    ''' <remarks></remarks>
    Public Shared Operator ^(x As BigInt, y As BigInt) As BigInt
        Dim ans As BigInt = x.Copy
        ans.Power(y, 0)
        Return ans
    End Operator

    ''' <summary>
    ''' Power (raise x to the power of y)
    ''' </summary>
    ''' <param name="x">The base value</param>
    ''' <param name="y">The Exponent</param>
    ''' <returns>A new BigInt with the results of the math operation</returns>
    ''' <remarks>Shared Function</remarks>
    Public Shared Function Power(x As BigInt, y As BigInt) As BigInt
        Dim ans As BigInt = x.Copy
        ans.Power(y, 0)
        Return ans
    End Function

    ''' <summary>
    ''' Power (raise x to the power of y) with required size
    ''' </summary>
    ''' <param name="x">The base value</param>
    ''' <param name="y">The Exponent</param>
    ''' <param name="size">The required size of the resultant BigInt</param>
    ''' <returns>A new BigInt with the results of the math operation</returns>
    ''' <remarks>Shared Function</remarks>
    Public Shared Function Power(x As BigInt, y As BigInt, size As Integer) As BigInt
        Dim ans As BigInt = x.Copy
        ans.Power(y, size)
        Return ans
    End Function

    ''' <summary>
    ''' Power (raise the current value to the power of y)
    ''' </summary>
    ''' <param name="y">The exponent</param>
    ''' <param name="size">The required bit size of the resultant value</param>
    ''' <remarks></remarks>
    Public Sub Power(ByVal y As BigInt, ByVal size As Integer)
        Dim real_bitsize As Integer
        Dim over As Boolean
        Dim temp_x, temp_y, one As BigInt

        temp_x = Me.Copy
        temp_y = y.Copy

        ' set to the correct size first, then add padding
        real_bitsize = If(size > 0, size, If(BitSize >= temp_y.BitSize, BitSize, temp_y.BitSize))
        SetSize(RoundUp(real_bitsize))
        temp_x.SetSize(BitSize)
        temp_y.SetSize(BitSize)
        one = BigInt.One(BitSize)

        ' There are a whole bunch of special cases!
        If temp_y.IsZero() Then
            ' Any number raised to the 0th power is 1
            Clear()
            n(0) = 1
            SetSize(real_bitsize)
            Exit Sub
        End If
        If temp_y.Equals(one) Then
            ' any number raised to 1 is itself
            SetSize(real_bitsize)
            Exit Sub
        End If
        If temp_x.Equals(one) Then
            ' any power of 1 is 1
            SetSize(real_bitsize)
            Exit Sub
        End If
        If IsZero() Or (temp_y.SignBit And (pfvArithMode <> ArithMode.Unsigned)) Then
            ' zero or fraction
            SetSize(real_bitsize)
            Clear()
            Exit Sub
        End If

        ' Just multiply x with itself that many times
        temp_y._Subtract(one)
        Do While Not temp_y.IsZero()
            Multiply(temp_x, BitSize)
            If pfvOverflow Then
                over = True
            End If
            temp_y._Subtract(one)
        Loop

        pfvOverflow = over Or MaxBit() >= real_bitsize

        SetSize(real_bitsize)
    End Sub

    ''' <summary>
    ''' Square Root of x (using Newton's method)
    ''' </summary>
    ''' <param name="x">The value to use to get the square root</param>
    ''' <returns></returns>
    ''' <remarks>No fractional parts</remarks>
    Public Shared Function SquareRoot(x As BigInt) As BigInt
        Dim ans As BigInt = x.Copy
        ans.SquareRoot(0)
        Return ans
    End Function

    ''' <summary>
    ''' Square Root of x (using Newton's method) with required size
    ''' </summary>
    ''' <param name="x">The value to use to get the square root</param>
    ''' <param name="size">The required size of the resultant BigInt</param>
    ''' <returns></returns>
    ''' <remarks>No fractional parts</remarks>
    Public Shared Function SquareRoot(x As BigInt, size As Integer) As BigInt
        Dim ans As BigInt = x.Copy
        ans.SquareRoot(size)
        Return ans
    End Function

    ''' <summary>
    ''' Square Root of current value (using Newton's method)
    ''' </summary>
    ''' <param name="size">The required size of the resultant value</param>
    Public Sub SquareRoot(ByVal size As Integer)
        Dim temp_x, r As BigInt
        Dim real_bitsize As Integer

        real_bitsize = If(size > 0, size, BitSize)
        SetSize(RoundUp(real_bitsize))

        ' a quick sanity check
        If IsZero() Or ((pfvArithMode <> ArithMode.Unsigned) And SignBit()) Then
            pfvOverflow = True
            Clear()
            Exit Sub
        End If

        temp_x = Me.Copy
        r = New BigInt(BitSize, pfvArithMode)

        r.n(0) = 1
        Do While IsGreater(r)
            r = Divide(temp_x, Me)
            ' Using right shift instead of division by 2
            _Add(r)
            RightShift(1, False, BitSize)
        Loop

        ' Carry on remainder (not a perfect square)
        r = Multiply(Me, Me)
        r._Subtract(temp_x)
        pfvCarryBit = Not r.IsZero()

        ' remove the wiggle room (if any)
        SetSize(real_bitsize)
    End Sub
#End Region

#Region "Bitwise Operators"

    ''' <summary>
    ''' Bitwise AND operator
    ''' </summary>
    ''' <param name="y">The left-hand value</param>
    ''' <param name="x">The right-hand value</param>
    ''' <returns>A new BigInt as the result of the Bitwise operation</returns>
    ''' <remarks></remarks>
    Public Shared Operator And(ByVal x As BigInt, ByVal y As BigInt) As BigInt
        Dim ans As BigInt = x.Copy
        ans.BitwiseAnd(y, x.BitSize)
        Return ans
    End Operator

    ''' <summary>
    ''' Perform a Bitwise AND operation of x and y
    ''' </summary>
    ''' <param name="y">The left-hand value</param>
    ''' <param name="x">The right-hand value</param>
    ''' <returns>A new BigInt as the result of the Bitwise operation</returns>
    ''' <remarks></remarks>
    Public Shared Function BitwiseAnd(ByVal x As BigInt, ByVal y As BigInt) As BigInt
        Dim ans As BigInt = x.Copy
        ans.BitwiseAnd(y, x.BitSize)
        Return ans
    End Function

    ''' <summary>
    ''' Perform a Bitwise AND operation of x and y with required size
    ''' </summary>
    ''' <param name="y">The left-hand value</param>
    ''' <param name="x">The right-hand value</param>
    ''' <param name="size">The required size of the resultant BigInt</param>
    ''' <returns>A new BigInt as the result of the Bitwise operation</returns>
    ''' <remarks></remarks>
    Public Shared Function BitwiseAnd(ByVal x As BigInt, ByVal y As BigInt, size As Integer) As BigInt
        Dim ans As BigInt = x.Copy
        ans.BitwiseAnd(y, size)
        Return ans
    End Function

    ''' <summary>
    ''' Perform a Bitwise AND operation
    ''' </summary>
    ''' <param name="y">The value for the Bitwise operation</param>
    ''' <param name="size">The required size of the resultant value</param>
    ''' <remarks></remarks>
    Public Sub BitwiseAnd(ByVal y As BigInt, ByVal size As Integer)
        Dim temp_y As BigInt

        ' First we have to make a copy of the variables
        temp_y = y.Copy

        ' pick the bit size of the answer
        SetSize(If(size > 0, size, If(BitSize >= temp_y.BitSize, BitSize, temp_y.BitSize)))

        ' Resize the y (since the size may have changed)
        temp_y.SetSize(BitSize)

        For i As Integer = 0 To LastByte
            Me.n(i) = Me.n(i) And temp_y.n(i)
        Next
    End Sub

    ''' <summary>
    ''' Bitwise OR operator
    ''' </summary>
    ''' <param name="y">The left-hand value</param>
    ''' <param name="x">The right-hand value</param>
    ''' <returns>A new BigInt as the result of the Bitwise operation</returns>
    ''' <remarks></remarks>
    Public Shared Operator Or(ByVal x As BigInt, ByVal y As BigInt) As BigInt
        Dim ans As BigInt = x.Copy
        ans.BitwiseOr(y, 0)
        Return ans
    End Operator

    ''' <summary>
    ''' Perform a Bitwise OR operation of x and y
    ''' </summary>
    ''' <param name="y">The left-hand value</param>
    ''' <param name="x">The right-hand value</param>
    ''' <returns>A new BigInt as the result of the Bitwise operation</returns>
    ''' <remarks>Shared function</remarks>
    Public Shared Function BitwiseOr(ByVal x As BigInt, ByVal y As BigInt) As BigInt
        Dim ans As BigInt = x.Copy
        ans.BitwiseOr(y, 0)
        Return ans
    End Function

    ''' <summary>
    ''' Perform a Bitwise OR operation of x and y with required size
    ''' </summary>
    ''' <param name="y">The left-hand value</param>
    ''' <param name="x">The right-hand value</param>
    ''' <returns>A new BigInt as the result of the Bitwise operation</returns>
    ''' <remarks>Shared function</remarks>
    Public Shared Function BitwiseOr(ByVal x As BigInt, ByVal y As BigInt, size As Integer) As BigInt
        Dim ans As BigInt = x.Copy
        ans.BitwiseOr(y, size)
        Return ans
    End Function

    ''' <summary>
    ''' Perform a Bitwise OR operation
    ''' </summary>
    ''' <param name="y">The value for the Bitwise operation</param>
    ''' <param name="size">The required size of the resultant value</param>
    ''' <remarks></remarks>
    Public Sub BitwiseOr(ByVal y As BigInt, ByVal size As Integer)
        Dim temp_y As BigInt

        temp_y = y.Copy

        SetSize(If(size > 0, size, If(BitSize >= temp_y.BitSize, BitSize, temp_y.BitSize)))

        temp_y.SetSize(BitSize)

        For i As Integer = 0 To LastByte
            Me.n(i) = Me.n(i) Or temp_y.n(i)
        Next
    End Sub

    ''' <summary>
    ''' Bitwise NOT operator of x
    ''' </summary>
    ''' <param name="x">The BigInt value</param>
    ''' <returns>A new BigInt as the result of the Bitwise operation</returns>
    ''' <remarks></remarks>
    Public Shared Operator Not(ByVal x As BigInt) As BigInt
        Dim ans As BigInt = x.Copy
        ans.OnesComplement(0)
        Return ans
    End Operator

    ''' <summary>
    ''' Perform a Bitwise NOT operation of x with required size
    ''' </summary>
    ''' <param name="x">The BigInt value</param>
    ''' <returns>A new BigInt as the result of the Bitwise operation</returns>
    ''' <remarks>Shared function</remarks>
    Public Shared Function OnesComplement(ByVal x As BigInt) As BigInt
        Dim ans As BigInt = x.Copy
        ans.OnesComplement(0)
        Return ans
    End Function

    ''' <summary>
    ''' Perform a Bitwise NOT operation
    ''' </summary>
    ''' <param name="size">The required size of the resultant value</param>
    ''' <remarks></remarks>
    Public Sub OnesComplement(ByVal size As Integer)
        If size > 0 Then
            SetSize(size)
        End If

        For i As Integer = 0 To LastByte
            Me.n(i) = CByte(255 - Me.n(i))
        Next

        Mask()
    End Sub

    ''' <summary>
    ''' Bitwise Exclusive OR operator
    ''' </summary>
    ''' <param name="y">The left-hand value</param>
    ''' <param name="x">The right-hand value</param>
    ''' <returns>A new BigInt as the result of the Bitwise operation</returns>
    ''' <remarks></remarks>
    Public Shared Operator Xor(ByVal x As BigInt, ByVal y As BigInt) As BigInt
        Dim ans As BigInt = x.Copy
        ans.[Xor](y, 0)
        Return ans
    End Operator

    ''' <summary>
    ''' Perform a Bitwise Exclusive OR operation on x and y
    ''' </summary>
    ''' <param name="y">The left-hand value</param>
    ''' <param name="x">The right-hand value</param>
    ''' <returns>A new BigInt as the result of the Bitwise operation</returns>
    ''' <remarks>Shared function</remarks>
    Public Shared Function [Xor](ByVal x As BigInt, ByVal y As BigInt) As BigInt
        Dim ans As BigInt = x.Copy
        ans.[Xor](y, 0)
        Return ans
    End Function

    ''' <summary>
    ''' Perform a Bitwise Exclusive OR operation on x and y with required size
    ''' </summary>
    ''' <param name="y">The left-hand value</param>
    ''' <param name="x">The right-hand value</param>
    ''' <returns>A new BigInt as the result of the Bitwise operation</returns>
    ''' <remarks>Shared function</remarks>
    Public Shared Function [Xor](ByVal x As BigInt, ByVal y As BigInt, size As Integer) As BigInt
        Dim ans As BigInt = x.Copy
        ans.[Xor](y, size)
        Return ans
    End Function

    ''' <summary>
    ''' Perform a Bitwise Exclusive OR operation
    ''' </summary>
    ''' <param name="y">The BigInt value</param>
    ''' <param name="size">The required size of the resultant value</param>
    ''' <remarks></remarks>
    Public Sub [Xor](ByVal y As BigInt, ByVal size As Integer)
        Dim temp_y As BigInt

        temp_y = y.Copy

        SetSize(If(size > 0, size, If(BitSize >= temp_y.BitSize, BitSize, temp_y.BitSize)))

        temp_y.SetSize(BitSize)

        For i As Integer = 0 To LastByte
            Me.n(i) = Me.n(i) Xor temp_y.n(i)
        Next
    End Sub

    ''' <summary>
    ''' Left Shift operator
    ''' </summary>
    ''' <param name="x">The value to be shifted</param>
    ''' <param name="distance">The number of bits to shift</param>
    ''' <returns>A BigInt value that contains the results of the operation</returns>
    ''' <remarks></remarks>
    Public Shared Operator <<(x As BigInt, distance As Integer) As BigInt
        Dim ans As BigInt = x.Copy
        ans.LeftShift(distance, 0)
        Return ans
    End Operator

    ''' <summary>
    ''' Shift the bit pattern of x to the Left
    ''' </summary>
    ''' <param name="x">The value to be shifted</param>
    ''' <param name="distance">The number of bits to shift</param>
    ''' <returns>A BigInt value that contains the results of the operation</returns>
    ''' <remarks>Shared Function</remarks>
    Public Shared Function LeftShift(x As BigInt, distance As Integer) As BigInt
        Dim ans As BigInt = x.Copy
        ans.LeftShift(distance, 0)
        Return ans
    End Function

    ''' <summary>
    ''' Shift the bit pattern of x to the Left with required size
    ''' </summary>
    ''' <param name="x">The value to be shifted</param>
    ''' <param name="distance">The number of bits to shift</param>
    ''' <returns>A BigInt value that contains the results of the operation</returns>
    ''' <remarks>Shared Function</remarks>
    Public Shared Function LeftShift(x As BigInt, distance As Integer, size As Integer) As BigInt
        Dim ans As BigInt = x.Copy
        ans.LeftShift(distance, size)
        Return ans
    End Function

    ''' <summary>
    ''' Shift the bit pattern to the Left
    ''' </summary>
    ''' <param name="distance">The number of bits to shift</param>
    ''' <param name="size">The required size of the resultant value</param>
    ''' <remarks></remarks>
    Public Sub LeftShift(distance As Integer, ByVal size As Integer)
        Dim temp, carry As Integer
        Dim temp_x As BigInt

        If size > 0 Then
            SetSize(size)
        End If

        ' a bit of sanity checking
        If distance <= 0 Then
            Exit Sub
        End If

        ' Shift left is the same as "multiply by 2"
        For j As Integer = 1 To distance
            temp_x = Me.Copy

            ' Carry is based upon the left most bit
            pfvCarryBit = SignBit()

            carry = 0
            For i As Integer = 0 To LastByte
                temp = (temp_x.n(i) * 2) + carry
                Me.n(i) = CByte(temp Mod 256)
                carry = temp \ 256
            Next
        Next

        Mask()
    End Sub

    ''' <summary>
    ''' Right Shift operator
    ''' </summary>
    ''' <param name="x">The value to be shifted</param>
    ''' <param name="distance">The number of bits to shift</param>
    ''' <returns>A BigInt value that contains the results of the operation</returns>
    ''' <remarks></remarks>
    Public Shared Operator >>(x As BigInt, distance As Integer) As BigInt
        Dim ans As BigInt = x.Copy
        ans.RightShift(distance, False, 0)
        Return ans
    End Operator

    ''' <summary>
    ''' Shift the bit pattern of x to the Right
    ''' </summary>
    ''' <param name="x">The value to be shifted</param>
    ''' <param name="distance">The number of bits to shift</param>
    ''' <returns>A BigInt value that contains the results of the operation</returns>
    ''' <remarks>Shared Function</remarks>
    Public Shared Function RightShift(x As BigInt, distance As Integer) As BigInt
        Dim ans As BigInt = x.Copy
        ans.RightShift(distance, False, 0)
        Return ans
    End Function

    ''' <summary>
    ''' Shift the bit pattern of x to the Right with required size
    ''' </summary>
    ''' <param name="x">The value to be shifted</param>
    ''' <param name="distance">The number of bits to shift</param>
    ''' <returns>A BigInt value that contains the results of the operation</returns>
    ''' <remarks>Shared Function</remarks>
    Public Shared Function RightShift(x As BigInt, distance As Integer, size As Integer) As BigInt
        Dim ans As BigInt = x.Copy
        ans.RightShift(distance, False, size)
        Return ans
    End Function

    ''' <summary>
    ''' Shift the bit pattern to the Right
    ''' </summary>
    ''' <param name="distance">The number of bits to shift</param>
    ''' <param name="SaveSign">Should the value of the sign bit be preserved</param>
    ''' <param name="size">The required size of the resultant value</param>
    ''' <remarks></remarks>
    Public Sub RightShift(distance As Integer, SaveSign As Boolean, ByVal size As Integer)
        Dim carry As Integer
        Dim sign_bit As Boolean
        Dim temp_x As BigInt

        If size > 0 Then
            SetSize(size)
        End If

        ' a bit of sanity checking
        If distance <= 0 Then
            Exit Sub
        End If

        For j As Integer = 1 To distance
            temp_x = Me.Copy

            ' Carry is based upon the right most bit
            pfvCarryBit = CBool(temp_x.n(0) And 1)
            sign_bit = SignBit()

            ' Shift right is the same as "divide by 2"
            carry = 0
            For i As Integer = LastByte To 0 Step -1
                If i < LastByte Then
                    If (temp_x.n(i + 1) And 1) > 0 Then
                        carry = 128
                    Else
                        carry = 0
                    End If
                End If
                Me.n(i) = CByte((temp_x.n(i) \ 2) Or carry)
            Next

            If SaveSign And sign_bit And (pfvArithMode <> ArithMode.Unsigned) Then
                ' are we doing an Arithmetic Shift Right?
                BitSet(BitSize - 1)
            End If
        Next
    End Sub

    ''' <summary>
    ''' Rotate the bit pattern of x to the Right
    ''' </summary>
    ''' <param name="x">The value to be rotated</param>
    ''' <param name="distance">The number of bits to rotate</param>
    ''' <param name="WithCarry">Should the rotate logic include the carry bit</param>
    ''' <param name="StartingCarry">The starting carry flag setting</param>
    ''' <param name="size">The required size of the resultant value</param>
    ''' <returns>A BigInt value that contains the results of the operation</returns>
    ''' <remarks>The bit at position 0 is copied to the sign bit location.  Shared Function</remarks>
    Public Shared Function RightRotate(x As BigInt, distance As Integer, WithCarry As Boolean, StartingCarry As Boolean, size As Integer) As BigInt
        Dim ans As BigInt = x.Copy
        ans.RightRotate(distance, WithCarry, StartingCarry, size)
        Return ans
    End Function

    ''' <summary>
    ''' Rotate the bit pattern to the Right
    ''' </summary>
    ''' <param name="distance">The number of bits to rotate</param>
    ''' <param name="WithCarry">Should the rotate logic include the carry bit</param>
    ''' <param name="StartingCarry">The starting carry flag setting</param>
    ''' <param name="size">The required size of the resultant value</param>
    ''' <remarks>The bit at position 0 is copied to the sign bit location</remarks>
    Public Sub RightRotate(distance As Integer, WithCarry As Boolean, StartingCarry As Boolean, ByVal size As Integer)
        Dim carry As Integer
        Dim temp_x As BigInt
        ' v6.0.3 - 5 May 12

        If size > 0 Then
            SetSize(size)
        End If

        ' a bit of sanity checking
        If distance <= 0 Then
            Exit Sub
        End If

        For j As Integer = 1 To distance
            temp_x = Me.Copy

            ' Carry is based upon the right most bit
            pfvCarryBit = CBool(temp_x.n(0) And 1)

            ' Rotate right is similar to "divide by 2"
            carry = 0
            For i As Integer = LastByte To 0 Step -1
                If i < LastByte Then
                    If (temp_x.n(i + 1) And 1) > 0 Then
                        carry = 128
                    Else
                        carry = 0
                    End If
                End If
                Me.n(i) = CByte((temp_x.n(i) \ 2) Or carry)
            Next

            If WithCarry Then
                If StartingCarry Then
                    ' add the sign bit back in
                    BitSet(BitSize - 1)
                End If
            Else
                If CBool(temp_x.n(0) And 1) Then
                    ' add the bit at position 0 back in
                    BitSet(BitSize - 1)
                End If
            End If
            StartingCarry = pfvCarryBit
        Next
    End Sub

    ''' <summary>
    ''' Rotate the bit pattern of x to the Left
    ''' </summary>
    ''' <param name="x">The value to be rotated</param>
    ''' <param name="distance">The number of bits to rotate</param>
    ''' <param name="WithCarry">Should the rotate logic include the carry bit</param>
    ''' <param name="StartingCarry">The starting carry flag setting</param>
    ''' <param name="size">The required size of the resultant value</param>
    ''' <returns>A BigInt value that contains the results of the operation</returns>
    ''' <remarks>The sign bit is copied to position 0.  Shared Function</remarks>
    Public Shared Function LeftRotate(x As BigInt, distance As Integer, WithCarry As Boolean, StartingCarry As Boolean, size As Integer) As BigInt
        Dim ans As BigInt = x.Copy
        ans.LeftRotate(distance, WithCarry, StartingCarry, size)
        Return ans
    End Function

    ''' <summary>
    ''' Rotate the bit pattern to the Left
    ''' </summary>
    ''' <param name="distance">The number of bits to rotate</param>
    ''' <param name="WithCarry">Should the rotate logic include the carry bit</param>
    ''' <param name="StartingCarry">The starting carry flag setting</param>
    ''' <param name="size">The required size of the resultant value</param>
    ''' <remarks>The sign bit is copied to position 0</remarks>
    Public Sub LeftRotate(distance As Integer, WithCarry As Boolean, StartingCarry As Boolean, ByVal size As Integer)
        Dim temp, carry As Integer
        ' v6.0.3 - 5 May 12

        If size > 0 Then
            SetSize(size)
        End If

        ' a bit of sanity checking
        If distance <= 0 Then
            Exit Sub
        End If

        For j As Integer = 1 To distance
            ' Carry is based upon the left most bit
            pfvCarryBit = SignBit()

            ' Rotate left is similar to "multiply by 2"
            carry = 0
            For i As Integer = 0 To LastByte
                temp = (Me.n(i) * 2) + carry
                Me.n(i) = CByte(temp Mod 256)
                carry = temp \ 256
            Next

            If WithCarry Then
                If StartingCarry Then
                    ' put the carry bit at position 0
                    Me.BitSet(0)
                End If
            Else
                If pfvCarryBit Then
                    ' put the sign bit at position 0
                    Me.BitSet(0)
                End If
            End If
            StartingCarry = pfvCarryBit
        Next

        Mask()
    End Sub
#End Region

#Region "Equality Operators"

    ''' <summary>
    ''' Is the current BigInt equal to zero?
    ''' </summary>
    ''' <returns>True if BigInt is zero</returns>
    ''' <remarks></remarks>
    Public Function IsZero() As Boolean
        For i As Integer = 0 To LastByte
            If Me.n(i) <> 0 Then
                Return False
            End If
        Next

        ' UNDONE: In some situations you might want 1's comp mode 0 = -0
        Return True
    End Function

    ''' <summary>
    ''' The Equal operator
    ''' </summary>
    ''' <param name="x">left hand BigInt</param>
    ''' <param name="y">right hand BigInt</param>
    ''' <returns>True if the two arguments have equal values</returns>
    ''' <remarks></remarks>
    Public Shared Operator =(x As BigInt, y As BigInt) As Boolean
        If Compare(x, y) = 0 Then
            Return True
        End If
        Return False
    End Operator

    ''' <summary>
    ''' The Not Equal operator
    ''' </summary>
    ''' <param name="x">left hand BigInt</param>
    ''' <param name="y">right hand BigInt</param>
    ''' <returns>True if the two arguments do not have equal values</returns>
    ''' <remarks></remarks>
    Public Shared Operator <>(x As BigInt, y As BigInt) As Boolean
        If Compare(x, y) <> 0 Then
            Return True
        End If
        Return False
    End Operator

    ''' <summary>
    ''' Test for equality (is x = y)
    ''' </summary>
    ''' <param name="x">left hand BigInt</param>
    ''' <param name="y">right hand BigInt</param>
    ''' <returns>True if the two arguments have equal values</returns>
    ''' <remarks>Shared function</remarks>
    Public Overloads Shared Function Equals(ByVal x As BigInt, ByVal y As BigInt) As Boolean

        If (x Is Nothing) And (y Is Nothing) Then
            Return True
        End If

        If (x Is Nothing) Xor (y Is Nothing) Then
            Return False
        End If

        If Compare(x, y) = 0 Then
            Return True
        End If
        Return False
    End Function

    ''' <summary>
    ''' Test for equality
    ''' </summary>
    ''' <param name="y">right hand BigInt</param>
    ''' <returns>True if the arguments and current instance have equal values</returns>
    ''' <remarks></remarks>
    Public Overloads Function Equals(ByVal y As BigInt) As Boolean
        If Compare(Me, y) = 0 Then
            Return True
        End If
        Return False
    End Function


    ''' <summary>
    ''' Less Than operator
    ''' </summary>
    ''' <param name="x">left hand BigInt</param>
    ''' <param name="y">right hand BigInt</param>
    ''' <returns>True if the x is less than y</returns>
    ''' <remarks></remarks>
    Public Shared Operator <(x As BigInt, y As BigInt) As Boolean
        If Compare(x, y) = -1 Then
            Return True
        End If
        Return False
    End Operator

    ''' <summary>
    ''' Less Than
    ''' </summary>
    ''' <param name="y">right hand BigInt</param>
    ''' <returns>True if the current instance is less than y</returns>
    ''' <remarks></remarks>
    Public Function IsLess(ByVal y As BigInt) As Boolean
        If Compare(Me, y) = -1 Then
            Return True
        End If
        Return False
    End Function

    ''' <summary>
    ''' Less Than or Equal operator
    ''' </summary>
    ''' <param name="x">left hand BigInt</param>
    ''' <param name="y">right hand BigInt</param>
    ''' <returns>True if x &lt;= y</returns>
    ''' <remarks></remarks>
    Public Shared Operator <=(x As BigInt, y As BigInt) As Boolean
        If Compare(x, y) <= 0 Then
            Return True
        End If
        Return False
    End Operator

    ''' <summary>
    ''' Less Than or Equal
    ''' </summary>
    ''' <param name="y">right hand BigInt</param>
    ''' <returns>True if the current instance is &lt;= y</returns>
    ''' <remarks></remarks>
    Public Function IsLessOrEqual(ByVal y As BigInt) As Boolean
        If Compare(Me, y) <= 0 Then
            Return True
        End If
        Return False
    End Function

    ''' <summary>
    ''' Greater Than operator
    ''' </summary>
    ''' <param name="x">left hand BigInt</param>
    ''' <param name="y">right hand BigInt</param>
    ''' <returns>True if x &gt; y</returns>
    ''' <remarks></remarks>
    Public Shared Operator >(x As BigInt, y As BigInt) As Boolean
        If Compare(x, y) = 1 Then
            Return True
        End If
        Return False
    End Operator

    ''' <summary>
    ''' Greater Than
    ''' </summary>
    ''' <param name="y">right hand BigInt</param>
    ''' <returns>True if the current instance is &gt; y</returns>
    ''' <remarks></remarks>
    Public Function IsGreater(ByVal y As BigInt) As Boolean
        If Compare(Me, y) = 1 Then
            Return True
        End If
        Return False
    End Function

    ''' <summary>
    ''' Greater Than or Equal operator
    ''' </summary>
    ''' <param name="x">left hand BigInt</param>
    ''' <param name="y">right hand BigInt</param>
    ''' <returns>True if x &gt;= y</returns>
    ''' <remarks></remarks>
    Public Shared Operator >=(x As BigInt, y As BigInt) As Boolean
        If Compare(x, y) >= 0 Then
            Return True
        End If
        Return False
    End Operator

    ''' <summary>
    ''' Greater Than or Equal
    ''' </summary>
    ''' <param name="y">right hand BigInt</param>
    ''' <returns>True if the current instance is &gt;= y</returns>
    ''' <remarks></remarks>
    Public Function IsGreaterOrEqual(ByVal y As BigInt) As Boolean
        If Compare(Me, y) >= 0 Then
            Return True
        End If
        Return False
    End Function

    ''' <summary>
    ''' Compare a BigInt value with another
    ''' </summary>
    ''' <param name="y">the value to compare against</param>
    ''' <returns>Returns 1=greater than, -1=less than, and 0=equals</returns>
    ''' <remarks></remarks>
    Public Shared Function Compare(x As BigInt, y As BigInt) As Integer
        Dim temp_x, temp_y As BigInt

        ' make copies since we may change the bit size
        temp_x = x.Copy
        temp_y = y.Copy

        If temp_x.BitSize > temp_y.BitSize Then
            temp_y.SetSize(temp_x.BitSize)
        End If
        If temp_y.BitSize > temp_x.BitSize Then
            temp_x.SetSize(temp_y.BitSize)
        End If

        ' Comparing negative number is a bit easier, since we know that any
        ' positive number is greater than any negative number
        If temp_x.pfvArithMode <> ArithMode.Unsigned Then
            Dim sign_x As Boolean = temp_x.SignBit
            Dim sign_y As Boolean = temp_y.SignBit
            ' when either is negative (but not both)
            If sign_x Xor sign_y Then
                If sign_y Then
                    Return 1
                End If
                Return -1
            End If
            ' UNDONE: In some situations you might want 1's comp mode 0 = -0
        End If

        For i As Integer = temp_x.LastByte To 0 Step -1
            If temp_x.n(i) > temp_y.n(i) Then
                ' x > y
                Return 1
            End If
            If temp_x.n(i) < temp_y.n(i) Then
                ' x < y
                Return -1
            End If
        Next
        ' must be equal
        Return 0
    End Function

    ''' <summary>
    ''' Does the BigInt equate to True
    ''' </summary>
    ''' <param name="x">The BigInt value to test</param>
    ''' <returns>If BigInt does not equal 0</returns>
    ''' <remarks></remarks>
    Public Shared Operator IsTrue(ByVal x As BigInt) As Boolean
        Return Not x.IsZero()
    End Operator

    ''' <summary>
    ''' Does the BigInt equate to False
    ''' </summary>
    ''' <param name="x">The BigInt value to test</param>
    ''' <returns>If BigInt equals 0</returns>
    ''' <remarks></remarks>
    Public Shared Operator IsFalse(ByVal x As BigInt) As Boolean
        Return x.IsZero()
    End Operator

    ''' <summary>
    ''' Does the current instance equate to True
    ''' </summary>
    ''' <value></value>
    ''' <returns>If BigInt does not equal 0</returns>
    ''' <remarks></remarks>
    Public ReadOnly Property IsTrue As Boolean
        Get
            Return Not IsZero()
        End Get
    End Property
#End Region

#Region "Import/Export Functions"

    ''' <summary>
    ''' Convert to an integer
    ''' </summary>
    ''' <returns>The resultant value as an integer</returns>
    ''' <remarks>Sets Loss Of Precision</remarks>
    Public Function ToInteger() As Integer
        Dim x As BigInt

        x = Me.Copy
        x.SetSize(32)

        pfvLossOfPrecision = Not Equals(x)

        ' 2's complement mode rules apply!
        If Not BitConverter.IsLittleEndian Then
            Array.Reverse(x.n)
        End If

        Return BitConverter.ToInt32(x.n, 0)
    End Function

    ''' <summary>
    ''' Convert to a long
    ''' </summary>
    ''' <returns>The resultant value as an long</returns>
    ''' <remarks>Sets Loss Of Precision</remarks>
    Public Function ToLong() As Long
        Dim x As BigInt

        x = Me.Copy
        x.SetSize(64)
        pfvLossOfPrecision = Not Equals(x)

        ' 2's complement mode rules apply!
        If Not BitConverter.IsLittleEndian Then
            Array.Reverse(x.n)
        End If

        Return BitConverter.ToInt64(x.n, 0)
    End Function

    ''' <summary>
    ''' Convert to a double
    ''' </summary>
    ''' <returns>The resultant value as an double</returns>
    ''' <remarks>Sets Loss Of Precision</remarks>
    Public Function ToDouble() As Double
        ' To Double (actually from long)
        Return CDbl(Me.ToLong())
    End Function

    ''' <summary>
    ''' Assignment operator
    ''' </summary>
    ''' <param name="s">From a short</param>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public Shared Widening Operator CType(s As Short) As BigInt
        Dim ans As New BigInt(s, 16, ArithMode.TwosComp)
        Return ans
    End Operator

    ''' <summary>
    ''' Assignment operator
    ''' </summary>
    ''' <param name="i">From a integer</param>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public Shared Widening Operator CType(i As Integer) As BigInt
        Dim ans As New BigInt(i, 32, ArithMode.TwosComp)
        Return ans
    End Operator

    ''' <summary>
    ''' Assignment operator
    ''' </summary>
    ''' <param name="l">From a long</param>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public Shared Widening Operator CType(l As Long) As BigInt
        Dim ans As New BigInt(l, 64, ArithMode.TwosComp)
        Return ans
    End Operator

    ''' <summary>
    ''' Assignment operator
    ''' </summary>
    ''' <param name="d">From a double</param>
    ''' <returns></returns>
    ''' <remarks>Requires a cast since it is narrowing</remarks>
    Public Shared Narrowing Operator CType(d As Double) As BigInt
        Dim ans As New BigInt(d, 16, ArithMode.TwosComp)
        Return ans
    End Operator

    ''' <summary>
    ''' Override the default ToString() method
    ''' </summary>
    ''' <returns>A string representation of the BigInt value</returns>
    ''' <remarks></remarks>
    Public Overrides Function ToString() As String
        Return "&H" & ToStringHex()
    End Function

    ''' <summary>
    ''' To Hexadecimal string
    ''' </summary>
    ''' <returns>A string representation of the BigInt value</returns>
    ''' <remarks></remarks>
    Public Function ToStringHex() As String
        Dim j, ending As Integer
        Dim sb As New StringBuilder

        ' How many digits are possible 
        ending = (BitSize - 1) \ 4

        ' Instead of using the normal method, we use the GetAt function to
        ' extract the digit's value directly from the n() array.
        For i As Integer = 0 To ending
            ' Sneaky trick to prevent multiplication and division
            j = GetAt(i * 4, 4)
            If j > 9 Then
                ' UNDONE: This will probably break globalization
                sb.Insert(0, Chr(Asc("a") + j - 10))
            Else
                sb.Insert(0, CStr(j))
            End If
        Next

        Return sb.ToString
    End Function

    ''' <summary>
    ''' To Binary string
    ''' </summary>
    ''' <returns>A string representation of the BigInt value</returns>
    ''' <remarks></remarks>
    Public Function ToStringBin() As String
        Return ToStringBin(False)
    End Function

    ''' <summary>
    ''' To Binary string
    ''' </summary>
    ''' <param name="pad">Should the output contain spaces between bytes</param>
    ''' <returns>A string representation of the BigInt value</returns>
    ''' <remarks></remarks>
    Public Function ToStringBin(ByVal pad As Boolean) As String
        Dim sb As New StringBuilder
        Dim spaces As Integer = 0
        Dim ans As String

        For i As Integer = LastByte To 0 Step -1
            For j As Integer = 7 To 0 Step -1
                If (CByte(2 ^ j) And n(i)) > 0 Then
                    sb.Append("1")
                Else
                    sb.Append("0")
                End If
            Next
            If i <> 0 And pad = True Then
                sb.Append(" ")
                spaces += 1
            End If
        Next
        ' Trim to current word length, considering the optional spaces
        If pad Then
            ans = sb.ToString.Substring(sb.Length - (BitSize + spaces), BitSize + spaces)
        Else
            ans = sb.ToString.Substring(sb.Length - BitSize, BitSize)
        End If

        Return ans
    End Function

    ''' <summary>
    ''' To Octal String
    ''' </summary>
    ''' <returns>A string representation of the BigInt value</returns>
    ''' <remarks></remarks>
    Public Function ToStringOct() As String
        Dim j, ending As Integer
        Dim sb As New StringBuilder

        ' How many digits are possible
        ending = (BitSize - 1) \ 3

        ' Instead of using the normal method, we use the GetAt function to
        ' extract the digit's value directly from the n() array.
        For i As Integer = 0 To ending
            ' Sneaky trick to prevent multiplication and division
            j = GetAt(i * 3, 3)
            sb.Insert(0, CStr(j))
        Next

        Return sb.ToString
    End Function

    ''' <summary>
    ''' To Decimal String.
    ''' </summary>
    ''' <returns></returns>
    ''' <remarks>
    ''' This is the only type that has to deal with the "sign mode"... 
    ''' all others types (hex, oct, and bin) are inherently unsigned
    ''' </remarks>
    Public Function ToStringDec() As String
        Dim j, ending As Integer
        Dim tempx, ten, z As BigInt
        Dim is_neg As Boolean
        Dim sb As New StringBuilder

        tempx = Me.Copy
        tempx.SetSize(RoundUp(BitSize))
        ten = New BigInt(tempx.BitSize, pfvArithMode)
        is_neg = False

        Select Case pfvArithMode
            Case ArithMode.Unsigned
                ' Do nothing
            Case ArithMode.OnesComp
                If SignBit() Then
                    is_neg = True
                    tempx.OnesComplement(tempx.BitSize)
                End If
            Case ArithMode.TwosComp
                If SignBit() Then
                    is_neg = True
                    tempx.OnesComplement(tempx.BitSize)
                    tempx._Add(BigInt.One(tempx.BitSize))
                End If
        End Select

        ' How many decimal digits are possible
        ending = CInt(Math.Floor(Me.BitSize / 3.33333333)) + 1

        ' This normal stuff...  To get each digit, divide the number by 10
        ' and then get the remainder.  Repeat.
        ten.n(0) = 10
        For i As Integer = 1 To ending
            ' Can't use my sneaky GetAt function... dang!
            z = Remainder(tempx, ten)
            j = z.n(0)
            sb.Insert(0, CStr(j))
            tempx = Divide(tempx, ten)
        Next

        If is_neg Then
            sb.Insert(0, "-")
        End If

        Return sb.ToString
    End Function
#End Region

#Region "Misc Utilities"

    ''' <summary>
    ''' The number 1 as a Big Integer
    ''' </summary>
    ''' <param name="size">The size of the resultant Big Integer</param>
    ''' <returns></returns>
    ''' <remarks>The ArithMode is not really relevant</remarks>
    Public Shared Function One(size As Integer) As BigInt
        Dim ans As New BigInt(size, ArithMode.TwosComp)
        ans.n(0) = 1
        Return ans
    End Function

    ''' <summary>
    ''' Split a BigInt into two equal size words (even bit sizes only!)
    ''' </summary>
    ''' <remarks>The size of the 2 halves (called by reference) is determined
    '''  by the size of this instance</remarks>
    Public Function Split() As BigInt()
        Dim half_bitsize, half_maxbyte As Integer
        Dim Right_Hand, Left_Hand As BigInt
        Dim ans(2) As BigInt

        ' A wee bit of sanity checking
        If (BitSize Mod 2) > 0 Then
            Throw New ArgumentException("Bit size is not divisible by two")
        End If
        half_bitsize = CInt(BitSize / 2)
        half_maxbyte = (half_bitsize - 1) \ 8

        Right_Hand = New BigInt(half_bitsize, pfvArithMode)
        Left_Hand = New BigInt(half_bitsize, pfvArithMode)

        ' Right size (is easy)
        For i As Integer = 0 To half_maxbyte
            Right_Hand.n(i) = Me.n(i)
        Next
        Right_Hand.Mask()

        ' Left side (right shift "half_bitsize" number of times)
        Dim temp_x As BigInt = Me.Copy
        temp_x.RightShift(half_bitsize, False, BitSize)

        For i As Integer = 0 To half_maxbyte
            Left_Hand.n(i) = temp_x.n(i)
        Next
        Left_Hand.Mask()

        ans(0) = Right_Hand
        ans(1) = Left_Hand
        Return ans
    End Function

    ''' <summary>
    ''' Create a bit mask
    ''' </summary>
    ''' <param name="bits">The size of the mask in bits</param>
    ''' <param name="size">The required size of the BigInt</param>
    ''' <param name="LeftJustify">optionaly left justify the bit pattern</param>
    ''' <returns>A new bit mask</returns>
    ''' <remarks></remarks>
    Public Shared Function CreateMask(bits As Integer, LeftJustify As Boolean, size As Integer, mode As ArithMode) As BigInt
        Dim ans As New BigInt(size, mode)

        If LeftJustify Then
            For i As Integer = ans.BitSize - 1 To ans.BitSize - bits Step -1
                ans.BitSet(i)
            Next
        Else
            For i As Integer = 0 To bits - 1
                ans.BitSet(i)
            Next
        End If

        Return ans
    End Function

    ''' <summary>
    ''' Move the bit pattern as many spaces as needed to 
    ''' </summary>
    ''' <param name="x">The bit pattern</param>
    ''' <remarks>Writes directly to passed arguments (yuk!)</remarks>
    Public Shared Function LeftJustify(x As BigInt) As BigInt()
        Dim d As Integer
        Dim Justified, distance As BigInt
        Dim ans(2) As BigInt

        Justified = x.Copy
        d = x.BitSize - x.MaxBit - 1

        Justified.LeftShift(d, x.BitSize)
        distance = New BigInt(d, x.BitSize, x.pfvArithMode)

        ans(0) = Justified
        ans(1) = distance
        Return ans
    End Function

    ''' <summary>
    ''' Converts the number into its absolute value
    ''' </summary>
    ''' <param name="x"></param>
    ''' <returns></returns>
    ''' <remarks>Shared Function</remarks>
    Public Shared Function AbsoluteValue(x As BigInt) As BigInt
        Dim temp_x As BigInt = x.Copy
        temp_x.AbsoluteValue()
        Return temp_x
    End Function

    ''' <summary>
    ''' Converts the number into its absolute value
    ''' </summary>
    ''' <remarks>Uses 2's complement methodology</remarks>
    Public Sub AbsoluteValue()
        ' Is there anything to do?
        If Not Me.SignBit Then
            Exit Sub
        End If

        Select Case pfvArithMode
            Case ArithMode.Unsigned
                Exit Sub
            Case ArithMode.OnesComp
                OnesComplement(BitSize)
            Case ArithMode.TwosComp
                ' Check for largest possible negative number in order
                ' to set the Overflow flag
                Dim sign As BigInt = CreateMask(1, True, BitSize, ArithMode.Unsigned)
                If Equals(sign) Then
                    pfvOverflow = True
                    Exit Sub
                Else
                    _Subtract(BigInt.One(BitSize))
                    OnesComplement(BitSize)
                End If
        End Select
    End Sub

    ''' <summary>
    ''' Change the sign of the Big Integer
    ''' </summary>
    ''' <remarks>Sets the overflow flag if used in unsigned mode</remarks>
    Public Sub ChangeSign()

        If SignBit() Then
            ' is already negative, so go to positive
            Select Case pfvArithMode
                Case ArithMode.Unsigned
                    _Subtract(One(BitSize))
                    OnesComplement(BitSize)
                    ' Not really an error, but we do set the overflow flag
                    pfvOverflow = True
                Case ArithMode.OnesComp
                    OnesComplement(BitSize)
                Case ArithMode.TwosComp
                    _Subtract(One(BitSize))
                    OnesComplement(BitSize)
            End Select
        Else
            ' is already positive, so go to negative
            Select Case pfvArithMode
                Case ArithMode.Unsigned
                    OnesComplement(BitSize)
                    _Add(One(BitSize))
                    ' Not really an error, but we do set the overflow flag
                    pfvOverflow = True
                Case ArithMode.OnesComp
                    OnesComplement(BitSize)
                Case ArithMode.TwosComp
                    OnesComplement(BitSize)
                    _Add(One(BitSize))
            End Select
        End If
    End Sub

    ''' <summary>
    ''' Combine two halves into one value.
    ''' </summary>
    ''' <param name="Left_Hand">The Most significant word</param>
    ''' <param name="Right_Hand">The Least significant word</param>
    ''' <remarks>The size of the resulting value is determined by the size
    ''' of calling arguments.  Modifies the instance directly.</remarks>
    Public Sub Combine(ByVal Left_Hand As BigInt, ByVal Right_Hand As BigInt)
        Dim temp As BigInt
        Dim i As Integer

        ' A wee bit of sanity checking
        If (Left_Hand.BitSize <> Right_Hand.BitSize) Then
            Throw New ArgumentException("Bits sizes do not match")
        End If

        SetSize(Left_Hand.BitSize * 2)
        Clear()

        ' Right side
        For i = 0 To Right_Hand.MaxByte
            n(i) = Right_Hand.n(i)
        Next

        ' Left side (shift left "left_hand.bitsize" number of times)
        temp = Left_Hand.Copy
        temp.SetSize(Left_Hand.BitSize * 2)
        temp.LeftShift(Left_Hand.BitSize, 0)

        ' combine with the existing
        For i = 0 To temp.LastByte
            n(i) = n(i) Or temp.n(i)
        Next
    End Sub

    ''' <summary>
    ''' Count the number of set bits.
    ''' </summary>
    ''' <remarks>Who you calling a Sum Bit !?!?</remarks>
    Public Sub SumBits()
        Dim temp As BigInt
        Dim one As BigInt = BigInt.One(BitSize)

        temp = Me.Copy
        Clear()

        For i As Integer = LastByte To 0 Step -1
            For j As Integer = 7 To 0 Step -1
                If (temp.n(i) And CByte(2 ^ j)) > 0 Then
                    _Add(one)
                End If
            Next
        Next

    End Sub

    ''' <summary>
    ''' Clear the bit at the given location
    ''' </summary>
    ''' <param name="location">the location of the bit to be cleared</param>
    ''' <remarks>Does not considering BitSize</remarks>
    Public Sub BitClear(location As Integer)
        Dim n_byte As Integer
        Dim n_bit As Byte

        n_byte = location \ 8
        n_bit = Not CByte(2 ^ (location Mod 8))

        ' a quick sanity check.. no error message!
        If n_byte <= LastByte Then
            Me.n(n_byte) = Me.n(n_byte) And n_bit
        End If
    End Sub

    ''' <summary>
    ''' Set the bit at the given location
    ''' </summary>
    ''' <param name="location">the location of the bit to be set</param>
    ''' <remarks>Does not considering BitSize</remarks>
    Public Sub BitSet(location As Integer)
        Dim n_byte As Integer
        Dim n_bit As Byte

        n_byte = location \ 8
        n_bit = CByte(2 ^ (location Mod 8))

        ' a quick sanity check.. no error message!
        If n_byte <= LastByte Then
            Me.n(n_byte) = Me.n(n_byte) Or n_bit
        End If
    End Sub

    ''' <summary>
    ''' Test to see if bit at the given location is set or not
    ''' </summary>
    ''' <param name="location">the location of the bit to test</param>
    ''' <remarks>Does not considering BitSize</remarks>
    Public Function BitTest(location As Integer) As Boolean
        Dim n_byte As Integer
        Dim n_bit As Byte

        n_byte = location \ 8
        n_bit = CByte(2 ^ (location Mod 8))

        ' a quick sanity check.. no error message!
        If n_byte <= LastByte Then
            If (Me.n(n_byte) And n_bit) > 0 Then
                Return True
            End If
        End If
        Return False
    End Function
#End Region

#Region "Private Methods"
    ''' <summary>
    ''' Clone the BigInt value
    ''' </summary>
    ''' <returns>A new BigInt that is identical to the original</returns>
    ''' <remarks>Performs a "deep" copy</remarks>
    Private Function Copy() As BigInt
        Dim ans As BigInt

        ' start with copying reference types, this part is pretty easy
        ans = DirectCast(Me.MemberwiseClone(), BigInt)

        ' Now we make a copy of the array
        Dim n2(LastByte) As Byte
        For i As Integer = 0 To LastByte
            n2(i) = n(i)
        Next
        ans.n = n2

        Return ans
    End Function

    ''' <summary>
    ''' Set/Reset the bit size
    ''' </summary>
    ''' <param name="size">The new bit size</param>
    ''' <remarks>This does NOT trigger a LossOfPrecision</remarks>
    Public Sub SetSize(ByVal size As Integer)
        If size <> BitSize Then
            Dim temp As BigInt = Nothing

            ' If the size has grown, then we may have to fill in the
            ' extra bits with the sign extension.  If the size has
            ' shrunk, then you're on you own, bud.
            If (size > BitSize) And SignBit() And (pfvArithMode <> ArithMode.Unsigned) Then
                temp = CreateMask(size - BitSize, True, size, pfvArithMode)
            End If

            LastByte = (size - 1) \ 8
            ReDim Preserve n(LastByte)
            BitSize = size
            Mask()

            ' apply the extended sign mask
            If temp IsNot Nothing Then
                For i As Integer = 0 To LastByte
                    Me.n(i) = Me.n(i) Or temp.n(i)
                Next
            End If
        End If
    End Sub

    ''' <summary>
    ''' Add an extra byte for padding
    ''' </summary>
    ''' <param name="size"></param>
    ''' <remarks>Similar to SetSize, but does not do a sign extension</remarks>
    Private Sub AddPadding(size As Integer)
        LastByte = (size - 1) \ 8
        ReDim Preserve n(LastByte)
        BitSize = size
        Mask()
    End Sub

    ''' <summary>
    ''' Find the location of the most significant bit
    ''' </summary>
    ''' <returns>The bit location</returns>
    ''' <remarks></remarks>
    Private Function MaxBit() As Integer
        For i As Integer = LastByte To 0 Step -1
            For j As Integer = 7 To 0 Step -1
                If (n(i) And CByte(2 ^ j)) > 0 Then
                    Return (i * 8) + j
                End If
            Next
        Next
        ' Must be zero
        Return 0
    End Function

    ''' <summary>
    ''' Reset the BigInt to Zero
    ''' </summary>
    ''' <remarks>Only the bit size is retained</remarks>
    Private Sub Clear()
        For i As Integer = 0 To LastByte
            n(i) = 0
        Next

        ' Also reset the status flags
        pfvCarryBit = False
        pfvOverflow = False
        pfvLossOfPrecision = False
    End Sub

    ''' <summary>
    ''' Returns the location of the most significant digit (byte) in
    ''' the n() array
    ''' </summary>
    ''' <returns>the index of the byte</returns>
    ''' <remarks>Useful when doing math operators that can skip leading 0's</remarks>
    Private Function MaxByte() As Integer

        For i As Integer = LastByte To 0 Step -1
            If n(i) <> 0 Then
                Return i
            End If
        Next
        Return 0
    End Function

    ''' <summary>
    ''' Put the integer value at the specified starting bit location within
    ''' the n() array.  This prevents a lot of bit shifting or multiplication,
    ''' when you can place the value where you want it.
    ''' </summary>
    ''' <param name="i">The value to be inserted into the byte array</param>
    ''' <param name="Location">The location in the byte array</param>
    ''' <remarks>Used internally in multication and division</remarks>
    Private Sub PutAt(ByVal i As Integer, ByVal Location As Integer)
        Dim n_bit, n_byte As Integer
        Dim temp As Long

        n_byte = Location \ 8
        n_bit = Location Mod 8
        temp = CLng(i * (2 ^ n_bit))

        n(n_byte) = CByte(temp Mod 256)
        If temp > 255 Then
            If n_byte < LastByte Then
                n(n_byte + 1) = CByte(temp \ 256)
            Else
                pfvOverflow = True
            End If
        End If
    End Sub

    ''' <summary>
    ''' Get an integer value from the specified location (and length) within
    ''' the n() array.  Sorta like the Mid$ function for the n() array.
    ''' </summary>
    ''' <param name="Location">The starting location in the byte array</param>
    ''' <param name="Length">The number of bits to retrieve</param>
    ''' <returns>The value at the given location</returns>
    ''' <remarks>Used internally in multication and division</remarks>
    Private Function GetAt(ByVal Location As Integer, ByVal Length As Integer) As Integer
        Dim ans, temp, n_bit, n_byte As Integer

        n_byte = Location \ 8
        n_bit = Location Mod 8

        If n_byte < LastByte Then
            temp = CInt(n(n_byte)) + (CInt(n(n_byte + 1)) * 256)
        Else
            temp = CInt(n(n_byte))
        End If
        ans = CInt((temp \ CInt(2 ^ n_bit)) Mod (2 ^ Length))

        Return ans
    End Function

    ''' <summary>
    ''' Mask the current value to the proper number of bits 
    ''' </summary>
    ''' <remarks>Only tweaks the last byte</remarks>
    Private Sub Mask()
        Dim mask_value As Byte

        ' Anything to do?
        If BitSize = (LastByte + 1) * 8 Then
            Exit Sub
        End If

        mask_value = CByte(((2 ^ ((BitSize - 1) Mod 8)) * 2) - 1)

        ' Mask that last byte
        n(LastByte) = n(LastByte) And mask_value
    End Sub

    ''' <summary>
    ''' Check to see if the Sign Bit is set
    ''' </summary>
    ''' <returns>True if the sign bit is set</returns>
    ''' <remarks></remarks>
    Private Function SignBit() As Boolean
        Dim mask_value As Byte

        mask_value = CByte(2 ^ ((BitSize - 1) Mod 8))

        If (n(LastByte) And mask_value) > 0 Then
            Return True
        End If

        Return False
    End Function


    ''' <summary>
    ''' Round up the bitsize to the next byte and add 1 byte
    ''' </summary>
    ''' <param name="bit_size">The original bitsize</param>
    ''' <returns>The new rounded bitsize</returns>
    ''' <remarks></remarks>
    Private Shared Function RoundUp(bit_size As Integer) As Integer
        Return (((bit_size - 1) \ 8) + 2) * 8
    End Function

    ''' <summary>
    ''' Assign a value to a BigInt from a byte array
    ''' </summary>
    ''' <param name="bytes">The array of bytes that contains the data</param>
    ''' <param name="size">The required bit size</param>
    ''' <param name="negative">Is this a negative value?</param>
    ''' <remarks>Used internally in the constructors</remarks>
    Private Sub Import(bytes As Byte(), size As Integer, negative As Boolean)
        Dim last As Integer

        SetSize(size)
        last = If(LastByte > bytes.Length - 1, bytes.Length - 1, LastByte)

        For i As Integer = 0 To last
            Me.n(i) = bytes(i)
        Next

        ' Do we have any leftover bytes?
        pfvLossOfPrecision = False
        If last < bytes.Length - 1 Then
            For i As Integer = last + 1 To bytes.Length - 1
                If bytes(i) <> 0 Then
                    pfvLossOfPrecision = True
                    Exit For
                End If
            Next
        End If

        ' How about any leftover bits?
        If Not pfvLossOfPrecision Then
            pfvLossOfPrecision = MaxBit() > BitSize
        End If

        ' Fill in the missing sign extension
        If bytes.Length - 1 < LastByte And negative Then
            For i As Integer = bytes.Length To LastByte
                Me.n(i) = 255
            Next
        End If

        Mask()
    End Sub
#End Region
End Class
